# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for parsing and interacting with diagrams in a Capella model."""
from __future__ import annotations

__all__ = [
    "DiagramDescriptor",
    "enumerate_diagrams",
    "parse_diagrams",
    "parse_diagram",
]

import collections.abc as cabc
import pathlib
import typing as t
import urllib.parse

from lxml import etree

from capellambse import aird, helpers, loader

from . import _common as C
from . import _filters, _semantic, _visual


class DiagramDescriptor(t.NamedTuple):
    fragment: pathlib.Path
    name: str
    styleclass: str | None
    uid: str
    viewpoint: str
    target: etree._Element


def enumerate_diagrams(
    model: loader.MelodyLoader,
    *,
    err_abort: bool = False,
) -> cabc.Iterator[DiagramDescriptor]:
    """Enumerate the diagrams in the model.

    Parameters
    ----------
    model
        The MelodyLoader instance
    err_abort
        Abort when encountering any error.  If False, enumerate all
        diagrams that are usable and ignore the others.
    """
    raw_views = model.xpath2(C.XP_VIEWS)
    views: list[tuple[pathlib.PurePosixPath, etree._Element, str]] = []
    if len(raw_views) == 0:
        raise ValueError("Invalid XML: No viewpoints found")

    # Extract the views' names
    for i in raw_views:
        viewname = helpers.xpath_fetch_unique(
            "./viewpoint", i[1], "viewpoint description"
        )
        assert viewname is not None
        viewname = C.RE_VIEWPOINT.search(viewname.attrib["href"])
        if viewname is not None:
            viewname = viewname.group(1)
        views.append((*i, urllib.parse.unquote(viewname or "")))

    if not views:
        raise ValueError("No viewpoints found")

    descriptors: list[tuple[pathlib.PurePosixPath, etree._Element, str]] = []
    for view in views:
        descriptors += [
            (view[0], d, view[2]) for d in C.XP_DESCRIPTORS(view[1])
        ]

    if len(descriptors) == 0:
        raise ValueError("Invalid XML: No diagrams found")

    for descriptor in descriptors:
        name = uid = None
        try:
            name = descriptor[1].attrib["name"]
            uid = descriptor[1].attrib["repPath"]
            if not uid.startswith("#"):
                raise ValueError("Invalid diagram reference: {uid!r}")

            # Extract styleclass from diagram description
            styledescription = helpers.xpath_fetch_unique(
                "./description[@href]", descriptor[1], "style description"
            )
            assert styledescription is not None
            styledescription = styledescription.attrib["href"]
            # This style description is something resembling XPath, e.g.
            #     platform:/[...]#//[...]/@ownedRepresentations[name='$$$']
            # The thing we're interested in is denoted as $$$ above.
            styleclass_match = C.RE_STYLECLASS.search(styledescription)
            if styleclass_match is None:
                styleclass: str | None = None
            else:
                styleclass = urllib.parse.unquote(styleclass_match.group(1))

            target = helpers.fragment_link(
                descriptor[0],
                t.cast(
                    str,
                    helpers.xpath_fetch_unique(
                        "./target/@href", descriptor[1], "target href"
                    ),
                ),
            )

            yield DiagramDescriptor(
                fragment=pathlib.Path(descriptor[0]),
                name=name,
                styleclass=styleclass,
                uid=uid[1:],
                viewpoint=descriptor[2],
                target=target,
            )
        except Exception:
            C.LOGGER.exception(
                "Error parsing descriptor for diagram with uid %r and name %r",
                uid,
                name,
            )
            if err_abort:
                raise


def parse_diagrams(model: loader.MelodyLoader) -> cabc.Iterator[aird.Diagram]:
    """Parse all diagrams from the model."""
    for descriptor in enumerate_diagrams(model):
        C.LOGGER.info(
            "Extracting diagram %r of type %s",
            descriptor.name,
            descriptor.styleclass,
        )
        yield parse_diagram(model, descriptor)


def parse_diagram(
    model: loader.MelodyLoader, descriptor: DiagramDescriptor
) -> aird.Diagram:
    """Parse a single diagram from the model.

    Parameters
    ----------
    descriptor
        A DiagramDescriptor as obtained from :func:`enumerate_diagrams`.
    """
    diagram = aird.Diagram(
        descriptor.name, styleclass=descriptor.styleclass, uuid=descriptor.uid
    )
    dgtree = model[f"{descriptor.fragment}#{descriptor.uid}"]
    treedata = helpers.xpath_fetch_unique(
        C.XP_ANNOTATION_ENTRIES,
        dgtree,
        "ownedAnnotationsEntries with source GMF_DIAGRAMS",
        dgtree.attrib["uid"],
    )
    assert treedata is not None

    if treedata.attrib.get("measurementUnit", "Pixel") != "Pixel":
        C.LOGGER.warning(
            "Unknown measurement unit %r",
            treedata.attrib.get("measurementUnit"),
        )

    for data_elm in treedata.iterdescendants("children", "edges"):
        try:
            elm = _element_from_xml(
                C.ElementBuilder(
                    target_diagram=diagram,
                    diagram_tree=dgtree,
                    data_element=data_elm,
                    melodyloader=model,
                    fragment=descriptor.fragment,
                )
            )
        except C.SkipObject:
            continue
        assert elm is not None
        diagram.add_element(elm, False, force=True)

    if len(diagram) == 0:
        C.LOGGER.error(
            "Deserialized diagram %r is empty", descriptor.name or diagram.name
        )
    else:
        diagram.calculate_viewport()
        _filters.applyfilters(diagram, dgtree, model)
    return diagram


def _element_from_xml(ebd: C.ElementBuilder) -> aird.DiagramElement:
    """Construct a single diagram element from the model XML."""
    if ebd.data_element.get("element") is not None:
        factory = _semantic.from_xml
    else:
        factory = _visual.from_xml
    return factory(ebd)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implements a high-level interface to Capella projects."""
from __future__ import annotations

__all__ = ["MelodyModel"]

import logging
import os
import pathlib
import typing as t

from lxml import etree

import capellambse.helpers
import capellambse.pvmt
from capellambse import loader

from . import common, diagram  # isort:skip

# Architectural Layers
from .layers import oa, ctx, la, pa  # isort:skip

LOGGER = logging.getLogger(__name__)
XT_PROJECT = "org.polarsys.capella.core.data.capellamodeller:Project"
XT_LIBRARY = "org.polarsys.capella.core.data.capellamodeller:Library"
XT_SYSENG = "org.polarsys.capella.core.data.capellamodeller:SystemEngineering"


class MelodyModel:
    """Provides high-level access to a model.

    This class builds upon the lower-level
    :class:`capellambse.loader.core.MelodyLoader` to provide an
    abstract, high-level interface for easy access to various model
    aspects.
    """

    oa = common.ProxyAccessor(oa.OperationalAnalysis, rootelem=XT_SYSENG)
    sa = common.ProxyAccessor(ctx.SystemAnalysis, rootelem=XT_SYSENG)
    la = common.ProxyAccessor(la.LogicalArchitecture, rootelem=XT_SYSENG)
    pa = common.ProxyAccessor(pa.PhysicalArchitecture, rootelem=XT_SYSENG)
    diagrams = diagram.DiagramAccessor(
        None, cacheattr="_MelodyModel__diagram_cache"
    )

    _diagram_cache: loader.FileHandler
    _diagram_cache_subdir: pathlib.PurePosixPath

    def __init__(
        self,
        path: str | os.PathLike,
        *,
        diagram_cache: str | os.PathLike | None = None,
        diagram_cache_subdir: str | pathlib.PurePosixPath | None = None,
        **kwargs: t.Any,
    ) -> None:
        """Load a project from the filesystem.

        For complete information on which exact ``kwargs`` are
        supported, consult the documentation of the used file handler.
        Refer to the "See Also" section for a collection of links.

        Parameters
        ----------
        path
            Path or URL to the project. The following formats are
            accepted:

            * A path to a local ``.aird`` file.
            * A path to a local directory (requires ``entrypoint``).
            * A remote URL, with a protocol or prefix that indicates
              which file handler to invoke (requires ``entrypoint``).

              Examples:

              * ``git://git.example.com/model/coffeemaker.git``
              * ``git+https://git.example.com/model/coffeemaker.git``
              * ``git+ssh://git@git.example.com/model/coffeemaker.git``
        entrypoint
            Entrypoint from path to the main ``.aird`` file.
        revision
            Revision of Git repository
        disable_cache
            Disable caching (not supported by all filehandlers)
        update_cache
            Update the cache (not supported by all filehandlers)
        identity_file
            SSH Identity file (not supported by all filehandlers)
        known_hosts_file
            SSH Known Hosts file (not supported by all filehandlers)
        username
            username (not supported by all filehandlers)
        password
            password (not supported by all filehandlers)
        diagram_cache
            An optional place where to find pre-rendered, cached
            diagrams. When a diagram is found in this cache, it will be
            loaded from there instead of being rendered on access. Note
            that diagrams will only be loaded from there, but not be put
            back, i.e. to use it effectively, the cache has to be
            pre-populated.

            This argument accepts the same formats as ``path``.

            The file names looked up in the cache built in the format
            ``uuid.ext``, where ``uuid`` is the UUID of the diagram (as
            reported by ``diag_obj.uuid``) and ``ext`` is the render
            format. Example:

            - Diagram ID: ``_7FWu4KrxEeqOgqWuHJrXFA``
            - Render call: ``diag_obj.as_svg`` or ``diag_obj.render("svg")``
            - Cache file name: ``_7FWu4KrxEeqOgqWuHJrXFA.svg``
        diagram_cache_subdir
            A sub-directory prefix to prepend to diagram UUIDs before
            looking them up in the ``diagram_cache``.

        See Also
        --------
        capellambse.loader.filehandler.FileHandler :
            Abstract super class for file handlers. Contains information
            needed for implementing custom handlers.
        capellambse.loader.filehandler.localfilehandler.LocalFileHandler :
            The file handler responsible for local files and
            directories.
        capellambse.loader.filehandler.gitfilehandler.GitFileHandler :
            The file handler implementing the ``git://`` protocol.
        """
        self._loader = loader.MelodyLoader(path, **kwargs)
        self.info = self._loader.get_model_info()

        try:
            self._pvext = capellambse.pvmt.load_pvmt_from_model(self._loader)
        except ValueError as err:
            LOGGER.warning(
                "Cannot load PVMT extension: %s: %s", type(err).__name__, err
            )
            LOGGER.warning("Property values are not available in this model")
            self._pvext = None

        if diagram_cache:
            if diagram_cache == path:
                self._diagram_cache = self._loader.filehandler
            else:
                self._diagram_cache = loader.get_filehandler(diagram_cache)
            self._diagram_cache_subdir = pathlib.PurePosixPath(
                diagram_cache_subdir or "/"
            )

    @property
    def _element(self) -> etree._Element:
        for tree in self._loader.trees.values():
            if capellambse.helpers.xtype_of(tree.root) in {
                XT_PROJECT,
                XT_LIBRARY,
            }:
                return tree.root
        raise AttributeError("No viable root element found")

    @property
    def _model(self) -> MelodyModel:
        return self

    def save(self, **kw: t.Any) -> None:
        """Save the model back to where it was loaded from.

        Parameters
        ----------
        kw
            Additional keyword arguments accepted by the file handler in
            use. Please see the respective documentation for more info.

        See Also
        --------
        capellambse.loader.filehandler.localfilehandler.LocalFileHandler.write_transaction :
            Accepted ``**kw`` when using local directories
        capellambse.loader.filehandler.gitfilehandler.GitFileHandler.write_transaction :
            Accepted ``**kw`` when using ``git://`` and similar URLs
        """
        self._loader.save(**kw)

    def search(
        self, *xtypes: str | type[common.GenericElement]
    ) -> common.ElementList:
        r"""Search for all elements with any of the given ``xsi:type``\ s.

        If only one xtype is given, the return type will be
        :class:`common.ElementList`, otherwise it will be
        :class:`common.MixedElementList`.
        """
        xtypes_: list[str] = []
        for i in xtypes:
            if isinstance(i, type) and issubclass(i, common.GenericElement):
                xtypes_.append(common.build_xtype(i))
            elif ":" in i:
                xtypes_.append(i)
            else:
                for _, l in common.XTYPE_HANDLERS.items():
                    xtypes_.extend(t for t in l if t.endswith(":" + i))

        cls = (common.MixedElementList, common.ElementList)[len(xtypes) == 1]
        return cls(
            self,
            self._loader.find_by_xsi_type(*xtypes_),
            common.GenericElement,
        )

    def by_uuid(self, uuid: str) -> common.GenericElement:
        """Search the entire model for an element with the given UUID."""
        return common.GenericElement.from_model(self, self._loader[uuid])



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class ModelInfo:
    branch: str | None = None
    title: str | None = None
    url: str | None = None
    revision: str | None = None
    rev_hash: str | None = None
    capella_version: str | None = None

    def __post_init__(self) -> None:
        self.set_project_url()
        self.version_url = self.derive_version_url()

    def derive_version_url(self) -> str | None:
        if self.url and self.rev_hash:
            return "/".join(
                (
                    self.url,
                    "-",
                    "commit",
                    self.rev_hash,
                )
            )
        return None

    def set_project_url(self) -> None:
        """Set self.url to repository URL if the url is an HTTPS URL.
        Otherweise set it to None.
        """
        if self.url:
            if self.url.endswith(".git"):
                self.url = self.url[: -len(".git")]

            if not self.url.startswith("https://"):
                self.url = None

    def as_dict(self) -> dict[str, str | None]:
        return dataclasses.asdict(self)



# Copyright 2021 DB Netz AG
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tools for the Operational Analysis layer.

.. diagram:: [CDB] OA ORM
"""
import operator

from .. import common as c
from .. import crosslayer, diagram
from ..crosslayer import capellacommon, capellacore, cs, fa, interaction

XT_ARCH = "org.polarsys.capella.core.data.oa:OperationalAnalysis"

XT_EOCI = (
    "org.polarsys.capella.core.data.oa:EntityOperationalCapabilityInvolvement"
)


@c.xtype_handler(XT_ARCH)
class OperationalActivity(c.GenericElement):
    """An operational activity."""

    _xmltag = "ownedOperationalActivities"

    owning_entity = c.CustomAccessor(
        c.GenericElement,
        operator.attrgetter("_model.oa.all_entities"),
        matchtransform=operator.attrgetter("activities"),
    )


@c.xtype_handler(XT_ARCH)
class OperationalProcess(c.GenericElement):
    """An operational process."""

    _xmltag = "ownedFunctionalChains"

    involved = c.ProxyAccessor(
        c.GenericElement,
        fa.XT_FCI,
        aslist=c.MixedElementList,
        follow="involved",
    )


@c.xtype_handler(XT_ARCH)
class OperationalCapability(c.GenericElement):
    """A capability in the OperationalAnalysis layer."""

    _xmltag = "ownedOperationalCapabilities"

    involved_activities = c.ProxyAccessor(
        OperationalActivity,
        interaction.XT_CAP2ACT,
        aslist=c.ElementList,
        follow="involved",
    )
    involved_entities = c.ProxyAccessor(
        c.GenericElement,
        XT_EOCI,
        follow="involved",
        aslist=c.MixedElementList,
    )
    involved_processes = c.ProxyAccessor(
        OperationalProcess,
        interaction.XT_CAP2PROC,
        aslist=c.ElementList,
        follow="involved",
    )
    owned_processes = c.ProxyAccessor(OperationalProcess, aslist=c.ElementList)

    postcondition = c.AttrProxyAccessor(
        capellacore.Constraint, "postCondition"
    )
    precondition = c.AttrProxyAccessor(capellacore.Constraint, "preCondition")
    scenarios = c.ProxyAccessor(interaction.Scenario, aslist=c.ElementList)
    states = c.AttrProxyAccessor(
        capellacommon.State, "availableInStates", aslist=c.ElementList
    )

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class OperationalCapabilityPkg(c.GenericElement):
    """A package that holds operational capabilities."""

    _xmltag = "ownedAbstractCapabilityPkg"

    capabilities = c.ProxyAccessor(OperationalCapability, aslist=c.ElementList)

    packages: c.Accessor


class AbstractEntity(cs.Component):
    """Common code for Entities."""

    activities = c.ProxyAccessor(
        OperationalActivity,
        fa.XT_FCALLOC,
        aslist=c.ElementList,
        follow="targetElement",
    )
    capabilities = c.CustomAccessor(
        OperationalCapability,
        operator.attrgetter("_model.oa.all_capabilities"),
        matchtransform=operator.attrgetter("involved_entities"),
        aslist=c.ElementList,
    )


@c.xtype_handler(XT_ARCH)
class Entity(AbstractEntity):
    """An Entity in the OperationalAnalysis layer."""

    _xmltag = "ownedEntities"

    entities: c.Accessor


@c.xtype_handler(XT_ARCH)
class OperationalActivityPkg(c.GenericElement):
    """A package that holds operational entities."""

    _xmltag = "ownedFunctionPkg"

    activities = c.ProxyAccessor(OperationalActivity, aslist=c.ElementList)

    packages: c.Accessor


@c.xtype_handler(XT_ARCH)
class EntityPkg(c.GenericElement):
    """A package that holds operational entities."""

    _xmltag = "ownedEntityPkg"

    entities = c.ProxyAccessor(Entity, aslist=c.ElementList)
    state_machines = c.ProxyAccessor(
        capellacommon.StateMachine, aslist=c.ElementList
    )

    packages: c.Accessor


class OperationalAnalysis(crosslayer.BaseArchitectureLayer):
    """Provides access to the OperationalAnalysis layer of the model."""

    root_entity = c.ProxyAccessor(Entity, rootelem=EntityPkg)
    root_activity = c.ProxyAccessor(
        OperationalActivity, rootelem=OperationalActivityPkg
    )

    activity_package = c.ProxyAccessor(OperationalActivityPkg)
    capability_package = c.ProxyAccessor(OperationalCapabilityPkg)
    entity_package = c.ProxyAccessor(EntityPkg)

    all_activities = c.ProxyAccessor(
        OperationalActivity,
        aslist=c.ElementList,
        deep=True,
    )
    all_actors = c.CustomAccessor(  # type: ignore[misc]
        Entity,
        operator.attrgetter("all_entities"),
        elmmatcher=lambda x, _: x.is_actor,  # type: ignore[attr-defined]
        aslist=c.ElementList,
    )
    all_capabilities = c.ProxyAccessor(
        OperationalCapability,
        aslist=c.ElementList,
        deep=True,
    )
    all_entities = c.ProxyAccessor(
        Entity,
        aslist=c.ElementList,
        deep=True,
    )
    all_processes = c.ProxyAccessor(
        OperationalProcess,
        aslist=c.ElementList,
        deep=True,
    )
    diagrams = diagram.DiagramAccessor(
        "Operational Analysis", cacheattr="_MelodyModel__diagram_cache"
    )


c.set_accessor(
    OperationalCapability,
    "inheritance",
    c.ProxyAccessor(
        OperationalCapability,
        interaction.XT_CAP_GEN,
        follow="super",
        aslist=c.ElementList,
    ),
)
c.set_self_references(
    (OperationalActivityPkg, "packages"),
    (OperationalCapabilityPkg, "packages"),
    (Entity, "entities"),
    (EntityPkg, "packages"),
)
